/*

 ideserv.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdio.h>
#ifndef WIN32		//*NIX
#include <errno.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/resource.h>
#else			//WIN32
#include <errno.h>
#include "resource.h"
#include "timeval.h"
#include "sleep.h"
#define lstat(x,y) stat(x,y)
#define S_ISLNK(x) 0
#endif 			//ALL
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>
#include "my_getopt-1.4/getopt.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "driver.h"
#include "usb.h"
#ifndef WIN32
#include "fifo.h"
#endif
#include "rs232.h"
#include "x1541.h"
#include "pc64.h"

static const struct io_ops *io_ops;
FILE *lf=NULL;
struct t_buffer {
    unsigned char *data;
    unsigned short pointer;
    unsigned short length;
    FILE *file;
    int fd;
    int mode;
    int filesize;
} buff[16];
char errorbuff[256];

#define SERVER_PORT 64000
int sock;
unsigned char ebuf[1030];
struct sockaddr_in eserver;
struct sockaddr_in eclient;
unsigned int fromlen;

#include "crc8.h"
#include "ideserv.h"

#define CM_CLOSED 0
#define CM_DIR 1
#define CM_FILE 2
#define CM_COMPAT 3

#ifdef WIN32
/* Prototypes */
INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
long WINAPI PCLinkLoop(long);
/* Global variable */
NOTIFYICONDATA g_notifyIconData;
#endif

const char *argp_program_version="ideservd 0.26 20101228";

void convertfilename(char *s, char typek, char *vessz, char *name, char *type, unsigned char *mode) {
    int b,i,j;

    i=j=*vessz=*mode=0;

    for (b=0;s[b] && s[b]!='/';b++) {
        if (s[b]==':') {i=*vessz=j=*mode=0; continue;}

        switch (*vessz) {
        case 0:
            if (s[b]==typek) {
                (*vessz)++;
                break;
            }
            if (i<16) name[i++]=s[b];
            break;
        case 1:
            if (s[b]==',') {
                (*vessz)++;
                break;
            }
            if (j<3) type[j++]=s[b];
            break;
        case 2:if (mode) *mode=s[b];(*vessz)++;break;
        }
    }
    name[i]=0;
    type[j]=0;
    if (j==1)
        switch (type[0]) {
        case 'S':strcpy(type,"SEQ");break;
        case 'P':strcpy(type,"PRG");break;
        case 'U':strcpy(type,"USR");break;
        case 'B':strcpy(type,"DIR");break;
        case 'J':strcpy(type,"LNK");break;
        }
}

int resolv_path(char *s, char *keszpath, int resolv_all) {
    unsigned char mode;
    char lname[1000],name[17],type[4],vessz;
    struct dirent *ep;
    DIR *dp;
    int found;
    int fel=0,fel2=0;

    while (s[fel] && s[fel]!=':') fel++;
    fel=s[fel]?1:0;

    if (fel) {
        while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') s[fel2++]='/';
        fel=fel2+1;
    }

      if (s[fel]=='/') {
#ifndef WIN32
        keszpath[0]='/';
#else
        keszpath[0]='\\';
#endif

        keszpath[1]=0;
        found=1;
    } else {
        keszpath[0]=0;
        found=0;
    }
    fel=0;

    while (1) {

        while (s[fel] && (s[fel]=='/' || s[fel]==':')) fel++;

        if (!s[fel]) {
            if (!resolv_all) s[0]=0;
            return found;
        }

        if (!resolv_all) {
            fel2=fel;
            while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') fel2++;
            if (!s[fel2]) {
                fel2=fel;fel=0;
                while (s[fel2]) s[fel++]=s[fel2++];
                s[fel]=0;
                return 1;
            }
        }

        dp=opendir(keszpath[0]?keszpath:".");
        if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return 0;}
        convertfilename(s+fel,'/',&vessz,name,type,&mode);

        while ((ep=readdir(dp)))
        {
            struct stat buf;
            DIRECTORY_ENTRY_T dirent;
            char lname2[2020];

            strcpy(lname2,keszpath);
            strncpy(lname,ep->d_name,999);
            strncat(lname2,ep->d_name,999);

            if (lstat(lname2,&buf)) continue;

            createdirentry(ep,&buf,&dirent);

            if ((dirent.attrib & (A_DIR | A_CLOSED))!=(A_CLOSED | A_DIR)) continue;

            dirent.filesize[0]=0;
            if (!matchname(dirent.name,name)) continue;

            found=1;
            break;
        }
        closedir(dp);

        if (!found)
        {
            int f=0,g=fel;
            while (s[g]) {
                lname[f]=c64toascii[(unsigned char)s[g]];
                f++;g++;
            }
            lname[f]=0;
        }

        if ((strlen(lname)+strlen(keszpath)+2)>1000) return 0;

        strcat(keszpath,lname);
#ifndef WIN32
        strcat(keszpath,"/");
#else
        strcat(keszpath,"\\");
#endif
        while (s[fel] && s[fel]!='/' && s[fel]!=':') fel++;

        if (!found) return 0;
    }
}

void do_chdir(char *s) {
    char keszpath[1000];


    if (!resolv_path(s,keszpath,1))
    {
        if (!s[0]) seterror(34,0);
        else seterror(39,0);
        fprintf(lf,"No such directory called \"%s\".\n",keszpath);
        return;
    }
    fprintf(lf,"Chdir to \"%s\".\n",keszpath);
    errtochannel15(chdir(keszpath));
}

void do_mkdir(char *s) {
    unsigned char mode;
    struct dirent *ep;
    DIR *dp;
    char keszpath[1020],vessz,type[4],lname[1000];
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege2;}

    convertfilename(s,',',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;

        found=1;
        break;
    }

    closedir(dp);

    if (!found) {
        int f=0;
        while (s[f]) {
            if (f<16) {
                lname[f]=c64toascii[(unsigned char)s[f]];
                f++;
            }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
//#ifndef WIN32
        errtochannel15(mkdir(keszpath,0777));
//#else
//        errtochannel15(mkdir(keszpath));
//#endif
    }
vege2:
    if (!s[0]) {
        seterror(34,0);
    }
    fprintf(lf,"Mkdir \"%s\".\n",keszpath);
}

void do_scratch(char *s, int smode) {
    unsigned char mode;
    char lname[1000],type[4],vessz,keszpath[1000];
    struct dirent *ep;
    DIR *dp;
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege;}

    convertfilename(s,'=',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
    if (type[0]==0) strcpy(type,"*");

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        }
        else
            if ((dirent.attrib & A_ANY)==A_DIR) continue;

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        found=1;
        break;
    }

    closedir(dp);
    if (!found) seterror(1,0);
vege:
    if (!found) {
        int f=0,g=0;
        while (s[f]) {
            lname[f]=c64toascii[(unsigned char)s[f]];
            f++;
        }
        if (!smode) {
	    if (!relaxfiletypes || memcmp(type,"PRG",3)) {
        	lname[f++]=',';
        	while (type[g]) {
            	    lname[f]=c64toascii[(unsigned char)type[g]];
            	    f++;g++;
        	}
	    }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found) {
        if (smode)
            errtochannel15(rmdir(keszpath));
        else
            errtochannel15(unlink(keszpath));
    }

    if (!s[0]) {
        seterror(34,0);
    }
    if (smode)
        fprintf(lf,"Rmdir \"%s\"\n",keszpath);
    else
        fprintf(lf,"Rm \"%s\"\n",keszpath);
}

void commandchannel(char *s) {
    if (!strncmp(s,"CD:",3) || !strncmp(s,"CD/",3)) {
        s[2]=':';
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (!strncmp(s,"CD_",3)) {
        errtochannel15(chdir(".."));
        fprintf(lf,"Chdir \"..\"\n");
    } else if (!strncmp(s,"C\xd0",2)) {
        seterror(2,1);
        fprintf(lf,"Change to partition %d\n",s[2]);
    } else if (s[0]=='U' && ((s[1] & 0x0f)==9 || (s[1] & 0x0f)==10)) {
        seterror(73,0);
        fprintf(lf,"Identify device\n");
    } else if (!strncmp(s,"I",1)) {
        seterror(0,0);
        fprintf(lf,"Initialize\n");
    } else if (!strncmp(s,"V",1)) {
        seterror(0,0);
        fprintf(lf,"Validate\n");
    } else if (s[0]=='M' && s[1]=='-' && s[2]=='R') { 
        int i,j,k;
        j=s[5];if (j>50) j=50;
        k=s[3];
	for (i=0;i<j;i++) errorbuff[i]="IDE64 CARTRIDGE "[k++ & 15];
	errorbuff[i]=0;
        fprintf(lf,"Memory read $%04x %02x\n", (unsigned char)s[3] | ((unsigned char)s[4] << 8), (unsigned char)s[5]);
    } else if (s[0]) {
        seterror(31,0);
    } else fprintf(lf,"\n");
}

int statuserror(int usecrc) {
    char s[256];
    unsigned char fp;
    int r;

    if (arguments.mode != M_ETHERNET) {
        if (io_ops->getb() < 0) return 1;
    }
    fp = 0;
    while (fp<255) {
        if (arguments.mode == M_ETHERNET) {
            s[fp] = ebuf[fp + 2];
        } else {
            r = io_ops->getb();
            if (r < 0) return 1;
            s[fp] = r;
        }
        if (!s[fp]) break;
        fp++;
    }
    s[fp] = 0;

    if (check_trailer("status", usecrc)) return 1;

    alarm(0);

    fprintf(lf, "Command: \"%s\": ", s);
    commandchannel(s);

    if (arguments.mode == M_ETHERNET) {
        ebuf[0]=ebuf[1]=strlen(errorbuff);
        memcpy(ebuf+2,errorbuff,ebuf[0]);
        if (sendto(sock,ebuf,ebuf[0]+2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        crc_clear();
        fp = strlen(errorbuff);
        if (io_ops->sendb(fp)) return 1;
        if (io_ops->sendbytes(fp, (unsigned char *)errorbuff)) return 1;
        if (send_trailer(usecrc)) return 1;
    }
    return 0;
}

#include "compat.h"

int openfile(int usecrc) {
    char s[256];
    unsigned char status=0x80, mode='R';
    unsigned char channel,fp;
    unsigned long long l=0;
    unsigned int length=0;
    struct t_buffer *buffer;
    int r;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[1] & 0x0f;
    } else {
        r = io_ops->getb();
        if (r < 0) return 1;
        channel=r & 0x0f;
    }
    buffer=&buff[channel];

    fp=0;
    while (fp < 255) {
        if (arguments.mode == M_ETHERNET) {
            r = ebuf[fp+2];
        } else {
            r = io_ops->getb();
            if (r < 0) return 1;
        }
        s[fp] = r;
        if (!r) break;
        fp++;
    }
    s[fp]=0;

    if (check_trailer("open",usecrc)) return 1;

    alarm(0);

    {
        char vessz,type[4];
        
        if (buffer->file) {
            fclose(buffer->file);
            buffer->file=NULL;
        }
        buffer->mode=CM_CLOSED;

        if (s[0] == '$') {
            unsigned char dirline[32];
            struct dirent *ep;
            DIR *dp;
            int sum=0,partition;
            char keszpath[1000];
            buffer->mode=CM_DIR;
            buffer->length=buffer->pointer=0;

	    partition=!strncmp(&s[1],"=P",2);

            resolv_path(s+1,keszpath,0);
            fprintf(lf,"Open [%d]: \"$%s%s\"\n",channel,keszpath,s+1);
            convertfilename(s+1,'=',&vessz,s,type,&mode);

            if (s[0]==0) strcpy(s,"*");
            if (type[0]==0) strcpy(type,"*");

            if (channel==1) {
                status = 0x80 | 70;goto vege; //no channel
            } else if (channel>1) {
                dp=opendir(keszpath[0]?keszpath:".");
                if (!dp) {
                    fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);
                    status = 0x80 | 62;goto vege;
                }
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                buffer->data[buffer->length++]='I';

                memset(&dirline[0],32,16);
                memset(&dirline[16],0,8);
                dirline[24]=A_DIR;
                memcpy(&dirline[25],"DIR",3);
                memset(&dirline[28],0,4);
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,(DIRECTORY_ENTRY_T*)dirline);

                    for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                }

                errtochannel15(closedir(dp));
                status=0x80;//ok
            } else

            if (partition) {
                memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
                memcpy(&dirline[0x08],"IDESERVD 0.26   ",16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"   \42PARTITION 1\42      NET  \0",28);
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                errtochannel15(0);
                status=0x80;//ok
            } else {

                dp=opendir(keszpath[0]?keszpath:".");

                if (!dp) {
                    fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);
                    status = 0x80 | 62;goto vege;
                }

                memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
                memset(&dirline[0x08],32,16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    DIRECTORY_ENTRY_T dirent;
                    int offs;
                    unsigned int n;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,&dirent);

                    if (dirent.pointer[0] & 0x80) continue;
                    n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                    if (n>65535) n=65535;

                    dirent.filesize[0]=0;
                    if (!matchname(dirent.name,s)) continue;
                    dirent.time[0]=0;
                    if (!matchname(dirent.filetype,type)) continue;

                    dirline[0x00]=dirline[0x01]=1;
                    dirline[0x02]=n;
                    dirline[0x03]=n >> 8;
                    sum+=n;
                    memset(&dirline[0x04],32,27);
                    offs=7;
                    if (n>9) offs--;
                    if (n>99) offs--;
                    if (n>999) offs--;
                    dirline[offs++]=0x22;
                    n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                    dirline[offs++]=0x22;
                    offs+=(16-n);

                    dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                    n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                    if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                    for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                    if (sum>65535) sum=65535;
                }

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=sum;
                dirline[0x03]=sum >> 8;
                memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                errtochannel15(closedir(dp));
                status=0x80;//ok
            }
            length=buffer->length;mode='R';
        } else {
            char vessz,type[4];
            unsigned char overwrite;
            char lname[1000],keszpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;
            int b;

            overwrite=0;
            for (b=0;s[b];b++) if (s[b]==':') {
                if (s[0]=='@') overwrite=1;
                break;
            }

            resolv_path(s,keszpath,0);
            convertfilename(s,',',&vessz,s,type,&mode);if (!mode) mode='R';

            dp=opendir(keszpath[0]?keszpath:".");
            if (!dp) {
                fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);
                status = 0x80 | 62;goto vege;
            }
            if (type[0]==0) strcpy(type,(channel==1)?"PRG":"*");

            while ((ep=readdir(dp)))
            {
                struct stat buf;
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strcpy(lname2,keszpath);
                strncpy(lname,ep->d_name,999);
                strncat(lname2,ep->d_name,999);

                if (lstat(lname2,&buf)) continue;

                createdirentry(ep,&buf,&dirent);

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name,s)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;length=buf.st_size;
                break;
            }

            closedir(dp);

            if (!found) {
                int f=0,g=0;
                while (s[f]) {
                    if (f<16) {
                        lname[f]=c64toascii[(unsigned char)s[f]];
                        f++;
                    }
                }
		if (!relaxfiletypes || memcmp(type,"PRG",3)) {
            	    lname[f++]=',';
            	    while (type[g]) {
                	if (g<3) {
                    	    lname[f]=c64toascii[(unsigned char)type[g]];
                    	    f++;g++;
                	}
            	    }
		}
                lname[f]=0;
            }

            if (channel==0) mode='R';
            if (channel==1) mode='W';

            strcat(keszpath,lname);

            fprintf(lf,"Open [%d]: \"%s\" %c\n",channel,keszpath,mode);

            if (!s[0]) {
                status=0x80 | 34;goto vege;//no filename
            }

            if (mode=='W') {
                int f=0;
                length=0;

                while (lname[f]) {
                    if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                        status=0x80 | 33;goto vege;//illegal characters
                    }
                    f++;
                }

                if (found && !overwrite) {
                    status=0x80 | 63;//exists
                } else {
#ifndef WIN32
                    if (overwrite) buffer->fd=open(keszpath, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
                    else buffer->fd=open(keszpath, O_CREAT|O_EXCL|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
#else
                    if (overwrite) buffer->fd=open(keszpath, O_CREAT|O_RDWR|O_TRUNC|O_BINARY, S_IRUSR|S_IWUSR);
                    else buffer->fd=open(keszpath, O_CREAT|O_EXCL|O_RDWR|O_BINARY, S_IRUSR|S_IWUSR);
#endif
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=0x80;
                        buffer->file=fdopen(buffer->fd,"wb+");
                        buffer->mode=CM_FILE;
                        buffer->filesize=0;
                    }
                }
            } else {
                if (!found) {
                    status=0x80 | 62;//not found
                } else {
        	    if (mode=='R') {

#ifndef WIN32
                	buffer->fd=open(keszpath, O_RDONLY, 0);
#else
                	buffer->fd=open(keszpath, O_RDONLY|O_BINARY, 0);
#endif
                	if (buffer->fd<0) errtochannel15(1); else {
                	    status=0x80;//ok
                    	    buffer->file=fdopen(buffer->fd, "rb");
                    	    buffer->mode=CM_FILE;
                            buffer->filesize=length;
                	}
            	    }
        	    if (mode=='A' || mode=='M') {
#ifndef WIN32
                	buffer->fd=open(keszpath, O_RDWR, 0);
#else
                	buffer->fd=open(keszpath, O_RDWR|O_BINARY, 0);
#endif
                	if (buffer->fd<0) errtochannel15(1); else {
                	    status=0x80;//ok
                    	    buffer->file=fdopen(buffer->fd, "rb+");
                            buffer->mode=CM_FILE;
                            buffer->filesize=length;
                	}
            	    }
                }
            }
        }
    }
vege:
    if (arguments.mode == M_ETHERNET) {
        ebuf[0] = length;
        ebuf[1] = length >> 8;
        ebuf[2] = length >> 16;
        ebuf[3] = length >> 24;
        ebuf[4] = mode;
        ebuf[5] = status;
        if (sendto(sock,ebuf,6, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        crc_clear();
        if (io_ops->sendb(status)) return 1;
        if (io_ops->sendb(mode)) return 1;
        if (io_ops->sendb(length)) return 1;
        if (io_ops->sendb(length >> 8)) return 1;
        if (io_ops->sendb(length >> 16)) return 1;
        if (io_ops->sendb(length >> 24)) return 1;
        if (send_trailer(usecrc)) return 1;
    }
    return 0;
}

int closefile(int usecrc) {
    unsigned char channel;
    unsigned int length;
    struct t_buffer *buffer;
    int r;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[1] & 0x0f;
        length=ebuf[2];
        length|=ebuf[3] << 8;
        length|=ebuf[4] << 16;
        length|=ebuf[5] << 24;
    } else {
        r = io_ops->getb();if (r < 0) return 1;
        channel=r & 0x0f;
        r = io_ops->getb();if (r < 0) return 1;
        length = r;
        r = io_ops->getb();if (r < 0) return 1;
        length|=r << 8;
        r = io_ops->getb();if (r < 0) return 1;
        length|=r << 16;
        r = io_ops->getb();if (r < 0) return 1;
        length|=r << 24;
        if (check_trailer("close",usecrc)) return 1;
    }
    buffer=&buff[channel];
    fprintf(lf,"Close [%d] %d byte(s)\n",channel, length);
    if (buffer->file) {
        if (buffer->mode==CM_FILE && length>buffer->filesize) {
            fflush(buffer->file);
            ftruncate(buffer->fd,length);
        }
	fclose(buffer->file);
    }
    buffer->mode=CM_CLOSED;buffer->file=NULL;
    if (buffer->data) {
        free(buffer->data);
        buffer->data=NULL;
    }
    if (arguments.mode == M_ETHERNET) {
        ebuf[0]=ebuf[1]=0x80;
        if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;
    }
    return 0;
}

int readfile(int usecrc) {
    unsigned char channel, sectors;
    unsigned short int j;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct t_buffer *buffer;
    int fr;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[1] & 0x0f;
        buffer=&buff[channel];
        address=ebuf[2];
        address|=ebuf[3]*256;
        address|=ebuf[4]*65536;
        sectors=ebuf[5];
        if (sectors>2) sectors=2;
    } else {
        fr = io_ops->getb();if (fr < 0) return 1;
        channel=fr & 0x0f;
        buffer=&buff[channel];
        fr = io_ops->getb();if (fr < 0) return 1;
        address=fr;
        fr = io_ops->getb();if (fr < 0) return 1;
        address|=fr << 8;
        fr = io_ops->getb();if (fr < 0) return 1;
        address|=fr << 16;
        fr = io_ops->getb();if (fr < 0) return 1;
        sectors=fr;
        if (check_trailer("read",usecrc)) return 1;
        if (sectors>128) sectors=128;
    }
    fprintf(lf,"Read [%d] %06x %d sector(s)\n",channel,address, sectors);
    if (buffer->mode==CM_CLOSED) {
	fprintf(lf,"Channel not open %d\n",channel);
        if (arguments.mode == M_ETHERNET) {
            ebuf[0]=ebuf[1]=0x80 | 70;
            if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
        } else {
            crc_clear();
            if (io_ops->sendb(0x80 | 70)) return 1;
            if (send_trailer(usecrc)) return 1;
        }
        return 0;
    }
    if (buffer->mode!=CM_DIR && (buffer->mode!=CM_FILE || !buffer->file)) {
        fprintf(lf,"Not readable.\n");
        if (arguments.mode == M_ETHERNET) {
            ebuf[0]=ebuf[1]=0x80 | 70;
            if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
        } else {
            crc_clear();
            if (io_ops->sendb(0x80 | 70)) return 1;
            if (send_trailer(usecrc)) return 1;
        }
        return 0;
    }

    if (buffer->mode==CM_FILE) {
	unsigned long itt;
	fseek(buffer->file, address << 8, SEEK_SET);
        if (!buffer->data) {
            buffer->data=malloc(65536);
            if (!buffer->data) exit(1);//out of memory?!
        }
        itt=fread(buffer->data,1,sectors*512,buffer->file);

	if (itt<sectors*512) memset(buffer->data+itt,0,sectors*512-itt);
	j=0;
    } else {
        j=buffer->pointer;
    }

    if (arguments.mode == M_ETHERNET) {
        memcpy(ebuf,buffer->data+j,sectors*512);
        ebuf[sectors*512]=ebuf[1+sectors*512]=0x80;
        if (sendto(sock, ebuf, sectors*512+2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
        buffer->pointer+=sectors*512;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;
        gettimeofday(&start,NULL);
        while (sectors) {
            if (arguments.mode == M_RS232 || arguments.mode == M_RS232S) {
                fr = io_ops->getb();if (fr < 0) return 1;
                if (fr != 0x5a) {
                    fprintf(lf, "Frame error %02x\n", fr);
                    return 1;
                }
            }
            crc_clear();
            if (io_ops->sendbytes(512,buffer->data+j)) return 1;
            j+=512;
            if (io_ops->sendb(0x80)) return 1;
            if (send_trailer(usecrc)) return 1;
            sectors--;buffer->pointer=j;l+=512;
        }
        gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
            int ll=l;
            l=l*1000000/l2;
            fprintf(lf,"Read %d bytes at %lld bytes/seconds\n",ll,l);
        }
    }
    return 0;
}


int writefile(int usecrc) {
    unsigned char channel, sectors;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct t_buffer *buffer;
    int fr;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[1] & 0x0f;
        address=ebuf[2];
        address|=ebuf[3]*256;
        address|=ebuf[4]*65536;
        sectors=ebuf[5];
        if (sectors>2) sectors=2;
    } else {
        fr = io_ops->getb();if (fr < 0) return 1;
        channel=fr & 0x0f;
        fr = io_ops->getb();if (fr < 0) return 1;
        address=fr;
        fr = io_ops->getb();if (fr < 0) return 1;
        address|=fr << 8;
        fr = io_ops->getb();if (fr < 0) return 1;
        address|=fr << 16;
        fr = io_ops->getb();if (fr < 0) return 1;
        sectors=fr;
        if (check_trailer("write",usecrc)) return 1;
        if (sectors>128) sectors=128;
    }
    buffer=&buff[channel];
    fprintf(lf,"Write [%d] %06x %d sector(s)\n",channel,address,sectors);
    if (buffer->mode!=CM_FILE || !buffer->file) {
        fprintf(lf,"Not writeable.\n");
        if (arguments.mode == M_ETHERNET) {
            ebuf[0]=ebuf[1]=0x80 | 25;
            if (sendto(sock, ebuf, 2, 0,(struct sockaddr *)&eclient, fromlen)<0) return 1;
        } else {
            if (io_ops->sendb(0x80+25)) return 1;
            if (send_trailer(usecrc)) return 1;
        }
        return 0;
    }
    if (!buffer->data) {
        buffer->data=malloc(65536);
        if (!buffer->data) exit(1);//out of memory?!
    }
    fseek(buffer->file, address << 8, SEEK_SET);
    if (arguments.mode == M_ETHERNET) {
        if (fwrite(ebuf+6,1,512*sectors,buffer->file)==512*sectors) ebuf[0]=ebuf[1]=0x80; else ebuf[0]=ebuf[1]=0x80 | 25;
        if (sendto(sock, ebuf, 2, 0,(struct sockaddr *)&eclient, fromlen)<0) return 1;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;

        gettimeofday(&start,NULL);
        while (sectors) {
    	    crc_clear();
  	    if (io_ops->getbytes(512, buffer->data)) return 1;
            if (usecrc) {
    	        if (io_ops->getb() < 0) return 1;
    	        if (crc_get()) {
    	    	    seterror(23,1);
    	    	    fprintf(lf,"CRC error while write\n");
    	    	    io_ops->getb();
    	    	    return 1;
    	    	}
	    }
            fr = io_ops->getb();if (fr < 0) return 1;
            if (fr != 0x5a) {
                fprintf(lf, "Frame error %02x\n",fr);
                return 1;
            }
    	    crc_clear();
    	    if (fwrite(buffer->data,1,512,buffer->file)==512) {
                if (sectors==1 || arguments.mode == M_RS232 || arguments.mode == M_RS232S) {
		    if (io_ops->sendb(0x80)) return 1;
            	    if (send_trailer(usecrc)) return 1;
            	}
    	    } else {
	        if (io_ops->sendb(0x80+25)) return 1;
    	        if (send_trailer(usecrc)) return 1;
    	        break;
    	    }
            sectors--;
        }
	gettimeofday(&end,NULL);
        //l=bytes;
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
            int ll=l;
	    l=l*1000000/l2;
    	    fprintf(lf,"Wrote %d bytes at %lld bytes/seconds\n",ll,l);
        }
    }
    return 0;
}

#ifndef WIN32
void testarg(int argc,char *argv[]) {
    int c;

    for (;;) {
        static struct option long_options[]= {
            {"user", required_argument, 0, 'u'},
            {"group", required_argument, 0, 'g'},
            {"root", required_argument, 0, 'r'},
            {"log", required_argument, 0, 'l'},
            {"nice", required_argument, 0, 'n'},
            {"allprg", no_argument, 0, 'P'},
            {"help", no_argument, 0, '?'},
            {"usage", no_argument, 0, 1},
            {"version", no_argument, 0, 'V'},
            {"background", no_argument, 0, 'b'},
            {"hog", no_argument, 0, 'h'},
            {"device", required_argument, 0, 'd'},
            {"mode", required_argument, 0, 'm'},
            {"lptport", required_argument, 0, 'p'},
            {"ipaddress", required_argument, 0, 'i'},
            {"network", required_argument, 0, 'N'},
            {0,0,0,0}
        };
        int option_index=0;

        c=getopt_long(argc, argv,
                      "m:u:g:r:l:n:P?Vbhd:p:i:N:"
                      , long_options, &option_index);
        if (c==-1) {
            if (optind==argc)
                break;
            c='?';
        }

        switch (c) {
        case '?':
            printf(
                   "Usage: ideservd [OPTION...]\n"
                   "IDEDOS 0.91 PCLink fileserver\n"
                   "\n"
                   "  -b, --background\t     Fork into background\n"
                   "  -d, --device=DEVICE\t     Device (/dev/parport0 or /dev/ttyS0)\n"
                   "  -g, --group=GROUP\t     Group under we run (nogroup)\n"
                   "  -h, --hog\t\t     CPU hog (off)\n"
                   "  -i, --ipaddress=IP\t     IP address of C64 on network\n"
                   "  -l, --log=FILE\t     Logfile (stdout)\n"
                   "  -m, --mode=MODE\t     Mode (x1541, xe1541, xm1541, xa1541,\n"
                   "\t\t\t     pc64, pc64s, rs232, rs232s usb, fifo, eth)\n"
                   "  -n, --nice=ADJUST\t     Adjust nice level (-19)\n"
                   "  -N, --network=NUM\t     Network number (0)\n"
                   "  -p, --lptport=IOPORT\t     Printer port address (0x378)\n"
                   "  -P, --allprg\t\t     Almost everything to PRG\n"
                   "  -r, --root=DIRECTORY\t     Root directory (.)\n"
                   "  -u, --user=USER\t     User under we run (nobody)\n"
                   "  -?, --help\t\t     Give this help list\n"
                   "      --usage\t\t     Give a short usage message\n"
                   "  -V, --version\t\t     Print program version\n"
                   "\n"
                   "Mandatory or optional arguments to long options are also mandatory or optional\n"
                   "for any corresponding short options.\n"
                   "\n"
                   "Report bugs to <soci@c64.rulez.org>.\n"
                   );
                   exit(0);
        case 1:
            printf(
                   "Usage: ideservd [-bhP?V] [-d DEVICE] [-g GROUP] [-i IP] [-l FILE] [-m MODE]\n"
                   "            [-n ADJUST] [-N NUM] [-p IOPORT] [-r DIRECTORY] [-u USER]\n"
                   "            [--background] [--device DEVICE] [--group=GROUP] [--hog]\n"
                   "            [--ipaddress IP] [--log=FILE] [--mode MODE] [--nice=ADJUST]\n"
                   "            [--network NUM] [--lptport=IOPORT] [--allprg] [--root=DIRECTORY]\n"
                   "            [--user=USER] [--help] [--usage] [--version]\n"
                  );
            exit(0);
        case 'V':
            printf("%s\n",argp_program_version);
            exit(0);
        case 'b':arguments.background=1;break;
        case 'P':arguments.relaxfiletypes=1;break;
        case 'u':arguments.user=optarg;break;
        case 'g':arguments.group=optarg;break;
        case 'l':arguments.log=optarg;break;
        case 'r':arguments.root=optarg;break;
        case 'h':arguments.hog=1;break;
        case 'd':arguments.device=optarg;break;
        case 'm':
            if (!strcasecmp(optarg,"x1541")) arguments.mode=M_X1541;
            if (!strcasecmp(optarg,"xe1541")) arguments.mode=M_XE1541;
            if (!strcasecmp(optarg,"xm1541")) arguments.mode=M_XM1541;
            if (!strcasecmp(optarg,"xa1541")) arguments.mode=M_XA1541;
            if (!strcasecmp(optarg,"pc64")) arguments.mode=M_PC64;
            if (!strcasecmp(optarg,"pc64s")) arguments.mode=M_PC64S;
            if (!strcasecmp(optarg,"rs232")) arguments.mode=M_RS232;
            if (!strcasecmp(optarg,"rs232s")) arguments.mode=M_RS232S;
            if (!strcasecmp(optarg,"usb")) arguments.mode=M_USB;
            if (!strcasecmp(optarg,"fifo")) arguments.mode=M_FIFO;
            if (!strcasecmp(optarg,"eth")) arguments.mode=M_ETHERNET;
            break;
        case 'p':arguments.lptport=strtol(optarg,NULL,0) & 0xffff;break;
        case 'i':
            {
                struct hostent *hp;
                hp = gethostbyname(optarg);
                if (hp) memcpy(&arguments.sin_addr, hp->h_addr, sizeof(arguments.sin_addr));
            }
            break;
        case 'N':arguments.network=strtol(optarg,NULL,0) & 0xff;break;
        case 'n':arguments.priority=strtol(optarg,NULL,0);break;
        default:
            exit(1);
        }
    }
}
#else // WIN32
void testarg(int argc,char *argv[]) {
    int j;
    for (j=1;j<argc;j++)
    {
	if (!strcmp(argv[j],"-?") || !strcmp(argv[j],"--help")) {
                MessageBox (NULL,
		"Usage: ideservd [OPTION...]\n"
		"IDEDOS 0.91 PCLink fileserver\n"
                "\n"
                "  -d, --device\t     Serial port device (COM1)\n"
                "  -h, --hog\t\t     CPU hog (off)\n"
                "  -i, --ipaddress\t      IP address of C64 on network\n"
                "  -l, --log\t\t     Logfile (ideservd.log)\n"
                "  -m, --mode\t\t     Mode (x1541, xe1541, xm1541, xa1541,\n"
                "\t\t\t     pc64, pc64s, rs232, rs232s usb, eth)\n"
                "  -N, --network\t      Network number (0)\n"
                "  -p, --lptport\t     Printer port address (0x378)\n"
                "  -P, --allprg\t     Almost everything to PRG\n"
		"  -?, --help\t     Give this help list\n"
		"      --usage\t     Give a short usage message\n"
		"  -V, --version\t     Print program version\n"
		"\n"
		"Mandatory or optional arguments to long options are also mandatory or optional\n"
		"for any corresponding short options.\n"
		"\n"
		"Report bugs to <soci@c64.rulez.org>.\n"
		,"Ideservd usage",MB_OK);
	    exit(1);
	}
        if (!strcmp(argv[j],"--usage")) {
            MessageBox(NULL,"Usage: ideservd [-hP?V] [-d DEVICE] [-i IP] [-l FILE] [-N NUM] [-p IOPORT] [-m MODE]\n"
                       "\t    [--device DEVICE] [--hog] [--ipaddress IP] [--log FILE] [--mode MODE] [--network NUM]\n"
                       "\t    [--lptport IOPORT] [--allprg] [--help] [--usage] [--version]\n" ,"Ideservd usage",MB_OK);
	    exit(1);
	}
	if (!strcmp(argv[j],"-V") || !strcmp(argv[j],"--version")) {
	    MessageBox(NULL,argp_program_version,"Ideservd version",MB_OK);
	    exit(1);
	}
	if (!strcmp(argv[j],"-P") || !strcmp(argv[j],"--allprg")) {arguments.relaxfiletypes=1;continue;}
	if (!strcmp(argv[j],"-l") || !strcmp(argv[j],"--log")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.log=realloc(arguments.log, strlen(argv[j])+1);
	    strcpy(arguments.log,argv[j]);
	    continue;
        }
	if (!strcmp(argv[j],"-h") || !strcmp(argv[j],"--hog")) {arguments.hog=1;continue;}
	if (!strcmp(argv[j],"-p") || !strcmp(argv[j],"--lptport")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.lptport=strtol(argv[j],NULL,0) & 0xffff;
	    continue;
	}
	if (!strcmp(argv[j],"-d") || !strcmp(argv[j],"--device")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.device=realloc(arguments.device, strlen(argv[j])+1);
	    strcpy(arguments.device,argv[j]);
	    continue;
        }
	if (!strcmp(argv[j],"-i") || !strcmp(argv[j],"--ipaddress")) {
            struct hostent *hp;
	    j++;if (j>=argc) goto ide2;
            hp = gethostbyname(argv[j]);
            if (hp) memcpy(&arguments.sin_addr, hp->h_addr, sizeof(arguments.sin_addr));
	    continue;
	}
	if (!strcmp(argv[j],"-m") || !strcmp(argv[j],"--mode")) {
	    j++;if (j>=argc) goto ide2;

	    arguments.device=malloc(strlen(argv[j])+1);
	    strcpy(arguments.device,argv[j]);
//@@@@
            if (!strcasecmp(argv[j],"usb")) arguments.mode=M_USB;
            if (!strcasecmp(argv[j],"eth")) arguments.mode=M_ETHERNET;
            if (!strcasecmp(argv[j],"pc64")) arguments.mode=M_PC64;
            if (!strcasecmp(argv[j],"pc64s")) arguments.mode=M_PC64S;
            if (!strcasecmp(argv[j],"x1541")) arguments.mode=M_X1541;
            if (!strcasecmp(argv[j],"xe1541")) arguments.mode=M_XE1541;
            if (!strcasecmp(argv[j],"xm1541")) arguments.mode=M_XM1541;
            if (!strcasecmp(argv[j],"xa1541")) arguments.mode=M_XA1541;
            if (!strcasecmp(argv[j],"rs232")) arguments.mode=M_RS232;
            if (!strcasecmp(argv[j],"rs232s")) arguments.mode=M_RS232S;
	    continue;
        }
	if (!strcmp(argv[j],"-N") || !strcmp(argv[j],"--network")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.network=strtol(argv[j],NULL,0) & 0xff;
	    continue;
	}

/*        if (!strcmp(argv[j],"-n")) {
	    int i=0;
	    j++;if (j>=argc)*/if (0) {
		ide2:
		MessageBox(NULL,"Ideservd: option requires an argument!",NULL,MB_ICONERROR|MB_OK);
		goto ide;
	    }
/*
	    while (argv[j][i] && argv[j][i]!='=') i++;
            if (argv[j][i]=='=') {
                argv[j][i]=0;
                arguments.priority=strtol(argv[j],NULL,0);
            }
            continue;
        }*/
	MessageBox(NULL,"Usage: ideservd [OPTION...]\n",NULL,MB_OK);
	ide:
        MessageBox(NULL,"Try `ideservd --help' or `ideservd --usage' for more information.\n",NULL,MB_OK);
        if (arguments.log) free(arguments.log);
        if (arguments.device) free(arguments.device);
        exit(1);
    }
}
#endif

void init(int argc,char *argv[])
{
#ifdef WIN32
    arguments.log="ideservd.log";
#else
    struct passwd *user2;
    struct group *group2;
    arguments.log="/dev/stdout";
#endif
    arguments.mode=M_USB;
    arguments.hog=0;
    arguments.device=NULL;
    arguments.lptport=DEFAULT_IOPORT;
    arguments.sin_addr.s_addr=htonl(INADDR_ANY);
    arguments.background=0;
    arguments.user= "nobody";
    arguments.group= "nogroup";
    arguments.root= ".";
    arguments.priority=-19;
    arguments.relaxfiletypes=0;
    testarg(argc, argv);

    relaxfiletypes=arguments.relaxfiletypes;
    if (!(lf=fopen(arguments.log,"wt"))) {fprintf(stderr,"Cannot open logfile '%s'\n",arguments.log);exit(1);}

    switch (arguments.mode) {
    case M_X1541:
    case M_XE1541:
    case M_XM1541:
    case M_XA1541:
        {
            int i;
            io_ops=&x1541_io_ops;
            i=arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_XE1541) i|=0x40000;
            if (arguments.mode==M_XM1541) i|=0x80000;
            if (arguments.mode==M_XA1541) i|=0xc0000;
            if (io_ops->initialize(arguments.device, i)) exit(1);
        }
        break;
    case M_PC64:
    case M_PC64S:
        {
            int i;
            io_ops=&pc64_io_ops;
            i=arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_PC64S) i|=0x40000;
            if (io_ops->initialize(arguments.device, i)) exit(1);
        }
        break;
    case M_RS232:
    case M_RS232S:
        io_ops=&rs232_io_ops;
#ifdef WIN32
        if (!arguments.device) arguments.device = "COM1";
#else
        if (!arguments.device) arguments.device = "/dev/ttyS0";
#endif
        if (io_ops->initialize(arguments.device, arguments.mode == M_RS232S)) exit(1);
        break;
    case M_USB:
        io_ops=&usb_io_ops;
        if (io_ops->initialize(NULL, 0)) exit(1);	//@@@@@@
        break;
    case M_FIFO:
#ifdef WIN32
	exit(1);
#else
        io_ops=&fifo_io_ops;
        if (io_ops->initialize(NULL, 0)) exit(1);
        break;
#endif
    case M_ETHERNET:
        sock=socket(AF_INET, SOCK_DGRAM, 0);
        if (sock < 0) {fprintf(stderr,"Cannot open socket\n");exit(1);}
        memset(&eserver,0,sizeof(eserver));
        eserver.sin_family=AF_INET;
        eserver.sin_addr.s_addr=SERVER_IP;
        eserver.sin_port=htons(SERVER_PORT+arguments.network);
        //     setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &so_reuseaddr, sizeof(so_reuseaddr));
        if (bind(sock,(struct sockaddr *)&eserver,sizeof(eserver))<0) {fprintf(stderr,"Cannot bind\n");exit(1);}
        fromlen = sizeof(struct sockaddr_in);
        break;
    }

#ifndef WIN32
    nice(arguments.priority);
    if (!(user2=getpwnam(arguments.user))) {fprintf(stderr,"User %s does not exists.\n",arguments.user);exit(1);}
    if (!(group2=getgrnam(arguments.group))) {fprintf(stderr,"Group %s does not exists.\n",arguments.group);exit(1);}
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
    chdir(arguments.root);
    if (arguments.mode != M_FIFO) {
        if (chroot(arguments.root) || chdir("/")) {fprintf(stderr,"Changing root dir to '%s'\n",arguments.root);exit(1);}
        if (setgid(group2->gr_gid) || setuid(user2->pw_uid)) {fprintf(stderr,"Could not drop privileges.\n");exit(1);}
    }
    if (arguments.background) if (fork()) {fprintf(stderr,"Forking into background.\n");exit(0);}
    fclose(stdin);
    fclose(stderr);
    fclose(stdout);
#endif
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
/*  int i;
  HWND hwnd_select;

  const char *pclinkmodes[] = {
            "x1541",
            "xe1541",
            "xm1541",
            "xa1541",
            "pc64",
            "pc64s",
            "rs232",
            "rs232s",
            "usb",
            "eth" };*/

  switch(message) 
  {
//   case WM_INITDIALOG:
//     hwnd_select = GetDlgItem(hwnd, IDC_SELECT);
//     for(i=0; i<11; i++) SendMessage(hwnd_select, CB_ADDSTRING, 0, (LPCTSTR)pclinkmodes[i]);
//     SendMessage(hwnd_select, CB_SETCURSEL, 8, 0); // from 0
//     return TRUE;
//   break;

   case WM_COMMAND:
	if (LOWORD(wParam) == IDC_EXIT || LOWORD(wParam) == IDCANCEL)
	{
		PostQuitMessage(0);
		EndDialog(hwnd,TRUE);
		return TRUE;
        }

	if (LOWORD(wParam) == IDC_BUTTON)
	{
	  ShowWindow(hwnd, SW_HIDE);
	}
   break;

//   case WM_SYSCOMMAND:
//	if (LOWORD(wParam) == SC_MINIMIZE)
//	{
//	   ShowWindow(hwnd, SW_HIDE);
//	}
//   break;

   case WM_USER:
	if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN)
        {
	   ShowWindow(hwnd, SW_SHOW);
   	}
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
// int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    int b;
#ifndef WIN32
    int n;
#endif
    init(argc,argv);
    fprintf(lf,"%s\n",argp_program_version);
    if (arguments.mode != M_ETHERNET)
        fprintf(lf,"Using driver %s\n", io_ops->drivername());
    signal(SIGALRM,timeout);
    signal(SIGTERM,terminate);
    signal(SIGINT,terminate);
    for (b=0;b<16;b++) {buff[b].mode=CM_CLOSED;buff[b].data=NULL;buff[b].file=NULL;}

#ifdef WIN32
    HWND hwnd;
    MSG msg;

    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);
 
// TRAY ICON

    memset(&g_notifyIconData, 0, sizeof(NOTIFYICONDATA));
    g_notifyIconData.cbSize = sizeof(NOTIFYICONDATA);
    g_notifyIconData.hWnd = hwnd;
    g_notifyIconData.uID = 0;
    g_notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    g_notifyIconData.uCallbackMessage = WM_USER; 
    g_notifyIconData.hIcon = (HICON)LoadImage( NULL, TEXT("pclink.ico"), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
    strcpy(g_notifyIconData.szTip, "IDE64 PCLink");
    Shell_NotifyIcon(NIM_ADD, &g_notifyIconData);
    
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,NULL,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    Shell_NotifyIcon(NIM_DELETE, &g_notifyIconData);
    terminate();
    return 0;
}

long WINAPI PCLinkLoop(long lParam)
{
    int b;
    int n;
#endif //WIN32

start:
    fprintf(lf,"ready.\n");
start2:
    crc_clear();
    if (arguments.mode == M_ETHERNET) {
        n = recvfrom(sock,ebuf+2,sizeof(ebuf)-2,0,(struct sockaddr *)&eclient,&fromlen);
        ebuf[0]=ntohs(eclient.sin_port) >> 8;
        ebuf[1]=ntohs(eclient.sin_port);
        eclient.sin_addr.s_addr=arguments.sin_addr.s_addr;
        if (n < 0) exit(1);
        if (n<(sizeof(ebuf)-2)) ebuf[n+2]=0;
        b=ebuf[0];
    } else {
        b=io_ops->wait();
    }
    if (b<0) terminate();	//@@@@@

    if (arguments.mode == M_ETHERNET && b==0) {
        b=ebuf[2];
    }

    switch (b)
    {
//    case 'X':return 0;
    case 0x00:goto start2;
    case 'N':b=openfile(0);break;
    case 0xCE:b=openfile(1);break;
    case 'G':b=readfile(0);break;
    case 0xC7:b=readfile(1);break;
    case 'S':b=writefile(0);break;
    case 0xD3:b=writefile(1);break;
    case 'D':b=closefile(0);break;
    case 0xC4:b=closefile(1);break;
    case 'I':b=statuserror(0);break;
    case 0xC9:b=statuserror(1);break;
    case 'O':b=openfile_compat();break;
    case 'R':b=readfile_compat();break;
    case 'C':b=closefile_compat();break;
    default:fprintf(lf,"Unknown command %02X\n",b);
    }
    if (b && arguments.mode != M_ETHERNET) {
	    io_ops->clean();
    }
    goto start;
}
